from typing import List

from .model import Model
from .process import Process


class Runner:
    process_: Process
    models_: List[Model]

    def __init__(self, process, models):
        self.process_ = process
        self.models_ = models

    def __call__(self):
        while True:
            data = self.process_.recv(0)
            if data is None:
                break
            model, parameters = data
            result = self.models_[model].create(parameters)()
            self.process_.send(0, (model, parameters, result))



from typing import Any, Tuple

from mpi4py import MPI


class Process:
    comm_: MPI.Intracomm
    rank_: int
    size_: int

    def __init__(self, comm=None):
        self.comm_ = comm if comm is not None else MPI.COMM_WORLD
        self.rank_ = self.comm_.Get_rank()
        self.size_ = self.comm_.Get_size()

    def size(self):
        return self.size_

    def rank(self):
        return self.rank_

    def send(self, destination: int, data: Any):
        self.comm_.send(data, dest=destination)

    def recv(self, source: int) -> Any:
        return self.comm_.recv(source=source)

    def recv_any(self) -> Tuple[int, Any]:
        status = MPI.Status()
        self.comm_.Probe(status=status)
        return status.source, self.comm_.recv(source=status.source)



from typing import List

from .controller import Controller
from .model import Model
from .process import Process
from .runner import Runner


def main(process: Process, models: List[Model]):
    if process.rank() == 0:
        controller = Controller(process, models)
        controller()
    else:
        runner = Runner(process, models)
        runner()



from typing import Callable, Any, List, Tuple, Mapping


class Model:
    name_: str
    creator_: Callable[[Mapping[str, Any]], Any]
    parameters_: List[Tuple[str, List[Any]]]

    def __init__(self, name, creator):
        self.name_ = name
        self.creator_ = creator
        self.parameters_ = []

    def name(self):
        return self.name_

    def parametrize(self, name: str, values: List[Any]) -> 'Model':
        self.parameters_.append((name, values))
        return self

    def __call__(self):
        current_parameters = [0] * len(self.parameters_)

        def generate(i):
            if i == len(self.parameters_):
                yield current_parameters
            else:
                for selected in range(len(self.parameters_[i][1])):
                    current_parameters[i] = selected
                    yield from generate(i + 1)

        yield from generate(0)

    def values(self, parameters: List[int]):
        return {name: values[parameter]
                for (name, values), parameter
                in zip(self.parameters_, parameters)}

    def create(self, parameters: List[int]):
        return self.creator_(self.values(parameters))



from typing import Generator, Optional, Tuple, Any, List

from .model import Model
from .process import Process


class Controller:
    process_: Process
    models_: List[Model]
    generator_: Optional[Generator[Tuple[int, Any], None, None]]
    ended_: bool
    left_: int

    def __init__(self, process, models):
        self.process_ = process
        self.models_ = models
        self.generator_ = None
        self.left_ = 0

    def send_all(self):
        for i in range(1, self.process_.size()):
            try:
                data = next(self.generator_)
                self.left_ += 1
                self.process_.send(i, data)
            except StopIteration:
                self.generator_ = None
                break

    def quit(self):
        for i in range(1, self.process_.size()):
            self.process_.send(i, None)

    def __call__(self):
        def models():
            for i, model in enumerate(self.models_):
                for parameters in model():
                    yield i, parameters

        self.left_ = 0
        self.generator_ = models()
        self.send_all()
        results = []
        while self.left_ > 0:
            source, result = self.process_.recv_any()
            results.append(result)
            self.left_ -= 1
            if self.generator_:
                try:
                    data = next(self.generator_)
                    self.left_ += 1
                    self.process_.send(source, data)
                except StopIteration:
                    self.generator_ = None
        print('Finished, quitting runners...')
        self.quit()
        print('Top 3 runs:')
        results = sorted(results, key=lambda v: v[-1], reverse=True)
        for model, parameters, score in results[:3]:
            print(f'Score: {score}, parameters: '
                  + ', '.join(f'{key}={repr(value)}'
                              for key, value
                              in self.models_[model].values(parameters).items()))
